/* Marmote is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Marmote is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Marmote. If not, see <http://www.gnu.org/licenses/>.

Copyright 2015 Alain Jean-Marie, Jean-Michel Fourneau, Jean-Marc Vincent, Issam Rabhi */

#include "marmoteSet.h"
#include <stdlib.h>
#include <stdio.h>

// Standard constructor. Represents the empty set.
marmoteSet::marmoteSet()
{
  _isSimple = true;
  _isUnion = false;
  _isProduct = false;
  _cardinal = 0;
  _nbZones = 0;
  _nbDimensions = 0;
  _totNbDims = 0;
  _zone = (marmoteSet**)NULL;
  _dimension = (marmoteSet**)NULL;
  _zeroState = (int*)NULL;
  _stateBuffer = (int*)NULL;
  _dimOffset = (int*)NULL;
}

// constructor of composite sets from
marmoteSet::marmoteSet( marmoteSet **list, int nb, opType t )
{
  switch( t ) {
  case UNION:
    _nbZones = nb;
    _isSimple = false;
    _isUnion = true;
    _isProduct = false;
    _totNbDims = 0;

    for ( int i = 0; i < _nbZones; i++ ) {
      _zone[i] = list[i];
      if ( _zone[i]->totNbDims() > _totNbDims ) {
        _totNbDims = _zone[i]->totNbDims();
      }
    }

    // initialize "zero" state as zero state of the first subset
    _zeroState = (int*)malloc( _zone[0]->totNbDims()*sizeof(int) );
    _zone[0]->firstState( _zeroState );

    break;

  case PRODUCT:
    _nbDimensions = nb;
    _isSimple = false;
    _isUnion = false;
    _isProduct = true;

    _dimOffset = (int*)malloc( _nbDimensions*sizeof(int) );
    _dimOffset[0] = 0;
    for ( int i = 0; i < _nbDimensions; i++ ) {
      _dimension[i] = list[i];
      if ( i < _nbDimensions-1 ) {
        _dimOffset[i+1] = _dimOffset[i] + _dimension[i]->totNbDims();
      }
    }

    // initialize "zero" state with zero states in each dimension
    _zeroState = (int*)malloc( _totNbDims*sizeof(int) );
    for ( int i = 0; i < _nbDimensions; i++ ) {
      _dimension[i]->firstState( &(_zeroState[_dimOffset[i]]) );
    }

    break;

  case SIMPLE:
    fprintf( stderr, "Error: improper use of marmoteSet constructor with SIMPLE type.\n");
  }

  _cardinal = cardinal();

}

marmoteSet::~marmoteSet()
{
  if ( _isProduct ) {
    free( _dimension );
  }
  else if ( _isUnion ) {
    free( _zone );
  }
  else {
    // elementary structures free their private arrays themselves
  }
}

long int marmoteSet::cardinal()
{
  long int res;

  if ( _isProduct ) {
    res = 1;
    for ( int i = 0; i < _nbDimensions; i++ ) {
      res = res * _dimension[i]->cardinal();
    }
  }
  else if ( _isUnion ) {
    res = 0;
    for ( int i = 0; i < _nbZones; i++ ) {
      res = res + _zone[i]->cardinal();
    }
  }
  else {
    // must be atomic
    res = _cardinal;
  }

  return res;
}

void marmoteSet::enumerate()
{

  if ( _isProduct ) {
    
  }
  else if ( _isUnion ) {
    for ( int i = 0; i < _nbZones; i++ ) {
      _zone[i]->enumerate();
    }
  }
  else {
    // must be atomic
    // selfEnumerate();
  }

}

// test whether some state buffer is the zero state
bool marmoteSet::isZero(int* buffer)
{
  bool res = true;

  if ( _isProduct ) {
    int* ptr;
    for( int i = 0; res && ( i < _nbDimensions ); i++ ) {
      ptr = &(buffer[_dimOffset[i]]);
      res = _dimension[i]->isZero( ptr );
    }
  }
  else if ( _isUnion ) {
    // no idea to which subset to ask...
  }
  else {
    // set is elementary
    fprintf( stderr, "Error: in marmoteSet::isZero(): elementary sets must have their own method. Ignored.\n" );
  }

  return res;
}

// set, in place, some state buffer to the zero state
void marmoteSet::firstState(int* buffer)
{
  if ( _isProduct ) {
    int* ptr;
    for( int i = 0; i < _nbDimensions; i++ ) {
      ptr = &(buffer[_dimOffset[i]]);
      _dimension[i]->firstState( ptr );
    }
  }
  else if ( _isUnion ) {
    _dimension[0]->firstState(buffer);
  }
  else {
    // set is elementary
    fprintf( stderr, "Error: in marmoteSet::firstState(): elementary sets must have their own method. Ignored.\n" );
  }
}

// increment, in place, the state buffer
void marmoteSet::nextState(int* buffer)
{
  if ( _isProduct ) {
    int numDim = _nbDimensions-1;
    int* ptr;
    do {
      ptr = &(buffer[_dimOffset[numDim]]);
      _dimension[numDim]->nextState( ptr );
      numDim--;
    } while( ( numDim >= 0 ) && _dimension[numDim]->isZero( ptr ) );
  }
  else if ( _isUnion ) {
    // in which subset is this state?? no idea!!
  }
  else {
    // set is elementary
    fprintf( stderr, "Error: in marmoteSet::nextState(): elementary sets must have their own method. Ignored.\n" );
  }
}

// converts an index into a state buffer
void marmoteSet::decodeState(int index, int* buffer)
{
  if ( _isProduct ) {
    int numDim = _nbDimensions-1;
    int* ptr;
    do {
      ptr = &(buffer[_dimOffset[numDim]]);
      _dimension[numDim]->nextState( ptr );
      numDim--;
    } while( ( numDim >= 0 ) && _dimension[numDim]->isZero( ptr ) );
  }
  else if ( _isUnion ) {
    // in which subset is this state?? no idea!!
  }
  else {
    // set is elementary
    fprintf( stderr, "Error: in marmoteSet::decodeState(): elementary sets must have their own method. Ignored.\n" );
  }
}

// printing a state as a vector
void marmoteSet::printState(FILE* out, int* buffer)
{
  if ( _isProduct ) {
    fprintf( out, "(" );
    int* ptr;
    for ( int i = 0; i < _nbDimensions; i++ ) {
      ptr = &(buffer[_dimOffset[i]]);
      _dimension[i]->printState( out, ptr );
      if ( i != _nbDimensions-1 ) {
        fprintf( out, ";" );
      }
      fprintf( out, ")" );
    }
  }
  else if ( _isUnion ) {
    // in which subset is this state?? no idea!!
  }
  else {
    // set is elementary
    fprintf( stderr, "Error: in marmoteSet::printState(): elementary sets must have their own method. Ignored.\n" );
  }
}

// printing a state as a vector, from its index number
void marmoteSet::printState(FILE *out, int index)
{
  decodeState( index, _stateBuffer );
  printState( out, _stateBuffer );
}
