/* Marmote is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Marmote is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Marmote. If not, see <http://www.gnu.org/licenses/>.

Copyright 2015 Alain Jean-Marie, Jean-Michel Fourneau, Jean-Marc Vincent, Issam Rabhi */

#include "markovChain.h"
#include <stdlib.h>
#include <iostream>
#include <string.h>
#include <math.h>
#include <cstdlib>
#include <fstream>
#include "transitionStructure/sparseMatrix.h"
#include "Distribution/diracDistribution.h"

//#include "PSI/psi_alias.h"
//#include "PSI/psi_traj.h"

#include "libXborne.h"
//#include "libPsi.h"

//#include "XBORNE/generMarkov/generMarkov.c"
//#include "XBORNE/gthLD/gthLD.c"
//#include "XBORNE/SOR/SOR.c"



#define SPARSE		1
#define FULL		2


markovChain::markovChain( int sz, timeType t )
{
  _stateSpaceSize = sz;
  _type = t;
  _debug = false;
  _isAbstract = false;
  _abstractNbre = 0;
  _abstract = (string*)NULL;

  _generator = (transitionStructure*)NULL;
  printf("Setting generator to NULL\n");
  _initDistribution = (discreteDistribution*)NULL;

}


markovChain::markovChain( transitionStructure* tr )
{
  _stateSpaceSize = tr->size();
  _type = tr->type();
  _debug = false;
  _isAbstract = false;
  _abstractNbre = 0;
  _abstract = (string*)NULL;

  _generator = tr;
  printf("Setting generator to some generator with type %d\n",
     _generator->type() );
  _initDistribution = (discreteDistribution*)NULL;

}


markovChain::markovChain(string format, string param[], int nbreParam,
			 string modelName, bool isAbstract)
{

  setModelName(modelName);
  setFormat(format);
  _isAbstract = isAbstract;

  if( _isAbstract )
    {
      _generator = (transitionStructure*)NULL;

      // arbitrary initialization of _initDistribution
      _initDistribution = new diracDistribution( 0 );

      // ERS format
      if ( format == "Ers" ) {

	setAbstractNbre(2);
	string tmp[2] = { modelName, modelName + ".mcl" };
	setAbstract(tmp);

      }
      
      // PSI input files
      else if( format == "PSI" ) {

	// TO BE REWRITTEN !!
	setAbstractNbre(1);
	string tmp[1] = { modelName };
	setAbstract(tmp);

	setSizeType( modelName + ".marca" );

	/*
	string ext = param[0].substr(param[0].find('.'),param[0].length() 
	- param[0].find('.'));
	string fileOut= param[0].substr(0,param[0].length()-ext.length());

	if(ext==".simu")
	  {
	    _abstract[0]=param[0];
	    setAbstractNbre(1);
	  }
	else if(ext == ".hbf_r" || ext ==".hbf_c" || ext ==".marca")
	  {
	    precomputation(param[0],modelName,ext);
	    
	    _abstract[0]=modelName+".simu";
	    
	    setAbstractNbre(1);
	    
	  }
	else
	  {
	    //to do
	    cout<<"other extension";
	  }
	*/
      }
      
      // XBORNE C-language format
      else if ( format == "XbornePres" ) {

	setAbstractNbre(4);
	string tmp[4] = { modelName, "const.h", "var.h", "fun.c" };
	setAbstract(tmp);

      }

      // XBORNE matrix format
      else if ( format == "Xborne" ){

	setAbstractNbre(4);
	string tmp[4] = { modelName, modelName+".Rii", modelName+".sz",
			  modelName+".cd" };
	setAbstract(tmp);

      }

    }
  else {
    // if not abstract...
    _debug = true;

#define ABORT(x) { printf("Error in markovChain: %s. Generator not initialized.\n", x ); \
  if ( _generator != (transitionStructure*)NULL ) { delete _generator; }; _generator = (transitionStructure*)NULL; }
#define WARNING(x) { printf("Warning in markovChain: %s.\n", x ); }

    if ( format == "Ers" ) {

      setFormat("Ers");

  FILE		*input;
  bool		reward = false;
  bool		readOK = true;
  double	epsilon = 1e-6;
  char		message[128];
  char		token[128];

  if ( modelName.size() == 0 )
    input = stdin;
  else {
    string fileName = modelName + ".mcl";

    if ( NULL == ( input = fopen( fileName.c_str(), "r" ) ) ) {
      sprintf( message, "could not open input file %s", fileName.c_str() );
      ABORT( message );
    }

  /* determine chain type */
  fscanf( input, "%128s", token );
  if ( ( !strcmp( token, "discrete" ) )
       || ( !strcmp( token, "1" ) ) ) {
    _type = DISCRETE;
  } 
  else if ( ( !strcmp( token, "continuous" ) )
       || ( !strcmp( token, "2" ) ) ) {
    _type = CONTINUOUS;
  }
  else {
    ABORT("could not read chain type");
  }

  if ( _debug ) 
    printf("Found time type = %d\n", _type );

  /* find out if reward or not */
  fscanf( input, "%128s", token );
  if ( !strcmp( token, "reward" ) ) {
    reward = true;
    /* read next token */
    fscanf( input, "%128s", token );
  }

  /* determine size, or sparse indicator */
  int matrixType = SPARSE; // this is the default
  if ( !strcmp( token, "sparse" ) ) {
    matrixType = SPARSE;
    readOK = ( 1 == fscanf( input, "%d", &(_stateSpaceSize) ) );
  }
  else {
    matrixType = FULL;
    readOK = ( 1 == sscanf( token, "%d", &(_stateSpaceSize) ) );
  }
  
  if ( !readOK )
    ABORT("could not read chain size");

  if ( reward ) {
    WARNING( "Markov Reward processes not implemented. Ignored." );
  }

  switch( matrixType ) {
  case FULL:
    WARNING( "Full matrices not implemented. Ignored." );

  case SPARSE:
    _generator = new sparseMatrix( _stateSpaceSize );
  }

  bool allocOK = ( (sparseMatrix*)NULL != _generator );

  if ( !allocOK )
    ABORT("could not allocate memory for matrix");

  /* ignore for the time being
  if ( Matrix_Type == FULL ) {
    for ( i=0; readOK && ( i < _stateSpaceSize ); i++ ) {
      for ( j=0; readOK && ( j < _stateSpaceSize ); j++ ) {

	if ( ( _type == DISCRETE )
	     || ( ( _type == CONTINUOUS ) && ( i != j ) ) ) {
	  
	  readOK = Read_Full_Matrix_Entry( input, _generator, i, j );
	
	  if ( !readOK ) {
	    sprintf( message, "failed reading matrix element (%d,%d)", i, j );
	    ABORT( message );
	  } else if (Debug ) {
	    printf("Got entry (%d,%d) = %f\n",
		   i, j, Get_Sparse_Matrix_Entry( _generator, i, j ) );
	  }
	}
      }
    }
  }
  else if ( Matrix_Type == SPARSE ) { 
  */
  do {
    int row;
    int col;
    double value;
    if ( 3 == fscanf( input, "%d %d %lf", &row, &col, &value ) ) {
      _generator->setEntry( row, col, value );
    }
    else
      readOK = false;
	
    if ( readOK && reward ) {
      /* get "wr" token, then read reward */
      fscanf( input, "%s", token );
      if ( strcmp( token, "wr" ) ) {
	WARNING( "expecting \"wr\" before reward but got" );
	fprintf( stderr, "%s.\n", token );
	
	/* Reading reward from this token... */
	if ( 1 != sscanf( token, "%lf", &value ) ) {
	  snprintf( message, 128, 
		    "Error in markovChain: unexpected token %s.",
		    token );
	  ABORT( message );
	}
      }
      else if ( 1 != fscanf( input, "%lf", &value ) ) {
	snprintf( message, 128,
		  "Error in markovChain: could not read reward" );
	ABORT( message );
      }

      // Set_Sparse_Matrix_Entry( _Reward, row, col, value );
    }
  } while ( readOK );

  // skip stopping mark
  fscanf( input, "%s", token );

  // read initial state for the chain
  int initState;
  readOK = ( 1 == fscanf( input, "%d", &(initState) ) );
  if ( !readOK ) {
    ABORT("could not read initial state");
  }
  else {
    _initDistribution = new diracDistribution( initState );
  }
    
  /* now checking/adjusting the generator */
  for ( int i=0; readOK && ( i < _stateSpaceSize ); i++ ) {
    double sum = _generator->rowSum( i );
    
    switch( _type ) {
    case DISCRETE:
      if ( fabs( sum - 1.0 ) > epsilon ) {
	printf("Warning in markovChain: row %d does not sum up to 1.0. Dif = %f.\n",
	       i, sum - 1.0 );
      }
      break;
    case CONTINUOUS:
      _generator->setEntry( i, i, -sum );
    }
  }

}

    }
    else if ( format == "MARCA" ) {

      setFormat("MARCA");

      FILE		*input;
      char		message[128];

      if ( modelName.size() == 0 )
	input = stdin;
      else {
	string fileName = modelName + ".marca";

	if ( NULL == ( input = fopen( fileName.c_str(), "r" ) ) ) {
	  snprintf( message, 128,
		    "could not open input file %s", fileName.c_str() );
	  ABORT( message );
	}

	// determine chain dimensions
	int rowDim;
	int colDim;
	int nbTrans;

	if ( 3 != fscanf( input, "%d%d%d", &colDim, &rowDim, &nbTrans ) ) {
	  ABORT("could not read chain size");
	}

	bool allocOK = ( (sparseMatrix*)NULL != _generator );

	if ( !allocOK )
	  ABORT("could not allocate memory for matrix");

	// reading transition values
	bool readOK = true;
	bool hasNegative = false;
	for ( int i = 0; ( i < nbTrans ) && readOK; i++ ) {
	  int row;
	  int col;
	  double value;
	  if ( 3 == fscanf( input, "%d %d %lf", &row, &col, &value ) ) {
	    _generator->setEntry( row, col, value );
	    hasNegative = hasNegative || ( value < 0.0 );
	  }
	  else {
	    readOK = false;
	    char message[256];
	    snprintf( message, 256,
		      "incomplete transition list: %d read, %d expected",
		      i, nbTrans );
	    ABORT(message);
	  }
	}
	
	// initial state for the chain is 0 by convention
	int initState = 0;
	_initDistribution = new diracDistribution( initState );

	// determine chain type
	if ( !hasNegative ) {
	  _type = DISCRETE;
	} 
	else {
	  _type = CONTINUOUS;
	}
	if ( _debug ) 
	  printf("Found time type = %d\n", _type );

	fclose( input );
      }
    }
    else if ( format == "HBF" ) {

      // reader not implemented

    }
    else if ( format == "XbornePres" ) {

      // model is in C-language. Call the MC generator, then read result
      char *gmp= (char *)(modelName.c_str());
      // Assumption: files const.h and var.h are present --> CHECK
      // Must **recompile** the genereMarkov program/object, then execute it.
      // Static call to mainGenerMarkov(gmp) will not work

      printf("Error: generation of model '%s' from format '%s' not implemented.\n",
	     gmp, format.c_str() );
      // reader not implemented...

    }
    else if ( format == "Xborne" ) {

      // reader not implemented

    }
  }
}


markovChain::~markovChain()
{
  if ( (discreteDistribution*)NULL != _initDistribution ) {
    delete _initDistribution;
  }
  if ( (transitionStructure*)NULL != _generator ) {
    delete _generator;
  }
  /*
  if ( (Distribution*)NULL != _initDistribution ) {
    delete _initDistribution;
  }
  */
  if ( _abstractNbre > 0 ) {
    delete[] _abstract;
  }
}


void markovChain::setAbstract( string abstract[] )
{
  _abstract = new string[_abstractNbre];

  for(int i=0; i<_abstractNbre;i++)
    {
      _abstract[i] = abstract[i];
    }
}

Distribution* markovChain::stationaryDistributionGthLD()
{
  Distribution*	distRes;

  if ( _isAbstract ) {
    
    if ( format() == "XbornePres" ) {

      // have to convert to "Xborne". To be done

    }
    else if ( format() == "Xborne" ) {

      char *mgp=(char *)(modelName().c_str());
      char *ext=(char *)"Rii";

      // call to the GTH method. Instantiates the stationary distribution
      // and the size of the chain (before that, it is not known)
      long double *pi;
      mainGthLD( mgp, ext, &pi, &(_stateSpaceSize) );

      // initialize the values and the distributions
      double *values = (double*) malloc( _stateSpaceSize * sizeof(double) );
      double *probas = (double*) malloc( _stateSpaceSize * sizeof(double) );
      for ( int i = 0; i < _stateSpaceSize; i++ ) {
        values[i] = (double)i;
        probas[i] = (double)pi[i];
      }

      distRes = new discreteDistribution( _stateSpaceSize, values, probas );

      free( values );
      free( probas );
      // eliminate temporary probability vector
      free( pi );
    }
  }
  else {
    // have to write as Xborne then call method on new "abstract" MC
  }

  return distRes;
}


Distribution* markovChain::stationaryDistributionSOR()
{
  Distribution*	distRes;

  if ( _isAbstract ) {
    
    if ( format() == "XbornePres" ) {

      // have to convert to "Xborne". To be done

    }
    else if ( format() == "Xborne" ) {

      char *mgp=(char *)(modelName().c_str());

      // call to the SOR method. Instantiates the stationary distribution
      // and the size of the chain (before that, it is not known)
      double *pi;
      mainSOR( mgp, &pi, &(_stateSpaceSize), /* writeResult? false */ 0 );

      /* initialize the values and the distributions */
      double *values = (double*) malloc( _stateSpaceSize * sizeof(double) );
      for ( int i = 0; i < _stateSpaceSize; i++ ) {
	values[i] = (double)i;
      }

      distRes = new discreteDistribution( _stateSpaceSize, values, pi );

      free( pi );
      free( values );
    }
  }
  else {
    // have to write as Xborne then call method on new "abstract" MC
  }

  return distRes;
}

Distribution* markovChain::stationaryDistribution( bool progress )
{
  Distribution*	distRes;


  switch( _type )
  {
  case DISCRETE:
    distRes = stationaryDistributionDT( progress );
    break;

  case CONTINUOUS:
    distRes = stationaryDistributionCT( progress );
    break;

  default:
    printf("Error: unknown Chain type (%d). No solution.\n", _type );
  }


  return distRes;
}


Distribution* markovChain::stationaryDistributionCT( bool progress )
{
  markovChain* unif;
  Distribution* res;

  unif = uniformize();

  res = unif->stationaryDistribution( progress );

  delete unif;

  return res;
}

/**
 * @brief Copy a Markov Chain, including the generator and the initial distribution.
 * @author Alain Jean-Marie
 * @return a new Markov chain, uniformized version of the original
 */
markovChain *markovChain::copy()
{
  markovChain* res;

  res = new markovChain( _stateSpaceSize, _type );
  if ( (transitionStructure*)NULL != _generator ) {
    transitionStructure* tr = _generator->copy();
    res->setGenerator( tr );
  }

  discreteDistribution* init = _initDistribution->copy();
  res->setInitDistribution( init );

  return res;
}

markovChain* markovChain::uniformize() 
{
  markovChain* res;

  if ( _type == DISCRETE ) {
    fprintf( stderr, "Error: cannot uniformize matrix. Copy of the original returned.\n" );
    res = copy();
  }
  else {
    // continuous-time object

    if ( _generator == NULL ) {
      // have to create generator. But from what?
      // makeMarkovChain();
      fprintf( stderr, "Error: no generator. Null returned.\n" );
      res = (markovChain*)NULL;
    }
    else {
      transitionStructure* uGen = _generator->uniformize();

      // possibly, uniformization did not work: for instance with infinite
      // state spaces. 
      if ( (transitionStructure*)NULL != uGen ) {
        res = new markovChain( uGen );
      }
      else {
        fprintf( stderr, "Error: could not uniformize transition structure. Null eturned.\n" );
        res = (markovChain*)NULL;
      }
    }
  }

  return res;
}


Distribution* markovChain::stationaryDistributionDT( bool progress )
{
  bool		continueIt;
  double	*values;
  double	*pi;
  double	*pi_it;
  double	*tmp;
  discreteDistribution  *res;
  discreteDistribution  *d1;
  discreteDistribution  *d2;
  double	err;
  int		time;
  int		tMax = 1000;

  values = (double*) malloc( _stateSpaceSize * sizeof(double) );
  pi = (double*) calloc( _stateSpaceSize, sizeof(double) );
  pi_it = (double*) calloc( _stateSpaceSize, sizeof(double) );

  _generator->write( stdout, "Ers" );

  /* initialize Pi, the values and the distributions */
  for ( int i = 0; i < _stateSpaceSize; i++ ) {
    values[i] = (double)i;
  }
  pi[0] = 1.0;
  d1 = new discreteDistribution( _stateSpaceSize, values, pi );
  d2 = new discreteDistribution( _stateSpaceSize, values, pi_it );

  continueIt = true;

  for ( time=0; (time<tMax) && continueIt; time++ ) {

    _generator->evaluateMeasure( pi, pi_it );
    
    if ( _debug ) {
      printf("# Pi(%d) = ", time );
      for ( int i=0; i<_stateSpaceSize; i++ ) {
        printf("%8.6f ", pi_it[i] );
      }
      printf("\n");
    }

    // compute distance
    // err = d1->distanceL1( d2 );
    err = 0.0;
    for ( int i = 0; i < _stateSpaceSize; i++ ) {
      err += fabs( pi[i] - pi_it[i] );
    }

    if ( progress ) {
      fprintf( stderr, "Iteration #%3d: distance %10e\n", time, err );
      fflush( stderr );
    }
    if ( err < 1e-7 ) {
      continueIt = false;
    } 
    else {
      // swap arrays and distribution objects
      discreteDistribution* dTmp = d1; d1 = d2; d2 = dTmp;
      tmp = pi_it; pi_it = pi; pi = tmp;
    }
  }
  if ( time >= tMax ) {
    fprintf( stderr, "Warning: Maximum number of iterations reached. ");
    fprintf( stderr, "Distribution may be very imprecise.\n");
  }

  // creating result. The array that is shipped away is pi. Should not be freed
  res = new discreteDistribution( _stateSpaceSize, values, pi );

  /* cleanup */
  free( pi );
  free( pi_it );
  free( values );
  delete d1;
  delete d2;

  /* returning result */
  return( res );

}


simulationResult* markovChain::simulateChainDT( int tMax,
						bool stats, bool traj,
						bool trace )
{
  int		curState;
  double	*cumTime;
  int		*states;
  double	*dates; // are needed for a common interface to simulationResult
  simulationResult *res;
  discreteDistribution** transDis;

  // prepare transition distributions
  transDis = 
    (discreteDistribution**) malloc( _stateSpaceSize * sizeof(discreteDistribution*) );
  for ( int i = 0; i < _stateSpaceSize; i++ ) {
    transDis[i] = _generator->getTransDistrib( i );
  }

  res = new simulationResult( _stateSpaceSize, DISCRETE, stats );
  if ( traj ) {
    /* allocate memory for entire trajectory: from 0 to tMax inclusive */
    res->setTrajectory(true);
    res->setTrajectorySize(1+tMax);
    dates = (double*)malloc( (size_t)(1+tMax) * sizeof(double) );
    states = (int*)malloc( (size_t)(1+tMax) * sizeof(int) );
  }

  // Sample from initial distribution
  if ( (discreteDistribution*)NULL == _initDistribution ) {
    cerr << "Warning in simulateChainDT: no initial distribution. Dirac(0) assumed." << endl;
    _initDistribution = new diracDistribution(0);
  }
  curState = _initDistribution->sample();
  
  if ( stats ) {
    cumTime = (double*) calloc( (size_t)_stateSpaceSize, sizeof( double ) );
  }

  for ( int time = 0; time < tMax; time++ ) {
    if ( traj ) {
        if ( trace ) {
          printf("%6d %4d\n", time, curState );
        }

        dates[time] = (double)time;
        states[time] = curState;
    }
    if ( stats ) {
      cumTime[ curState ]++;
    }

    curState = transDis[curState]->sample();

  }

  /* adjust end of simulation */
  if ( traj ) {
      if ( trace ) {
          printf("%6d %4d\n", tMax, curState );
        }

      dates[tMax] = (double)tMax;
      states[tMax] = curState;

  }

  // create result
  if ( stats && ( tMax > 0 ) ) {
    double *proba = 
      (double*) calloc( (size_t)_stateSpaceSize, sizeof( double ) );

    for ( int j=0; j<_stateSpaceSize; j++ ) {
      proba[j] = (double)cumTime[j]/(double)tMax;
      if ( trace ) {
	printf("State %3d:   cum time = %5d   ( %8.4f%%)\n",
	       j, (int)rint(cumTime[j]), 100.0*proba[j] );
      }
    }

    // at the moment, there is no generic state space at this level of 
    // abstract Markov Chains. So create one for the purpose.
    double* stateSpace = (double*)malloc( _stateSpaceSize * sizeof(double) );
    for ( int i = 0; i < _stateSpaceSize; i++ ) {
      stateSpace[i] = i;
    }

    discreteDistribution *dis =
      new discreteDistribution( _stateSpaceSize, stateSpace, proba );
    res->distrib( dis );
    
    free( cumTime );
    free( proba );
  }

  return res;
}


simulationResult* markovChain::simulateChainCT( double tMax,
						bool stats, bool traj,
						bool withIncrements,
						bool trace )
{
  int		curState;
  double	time;
  double	adv;
  double	*cumTime;
  double	*proba;
  double	*dates;
  int		*states;
  double	*increments;
  simulationResult* res;
  int		maxTrajLength;
  int		curTrajLength = 0;
  discreteDistribution** transDis;

  // prepare transition distributions
  transDis = 
    (discreteDistribution**) malloc( _stateSpaceSize * sizeof(discreteDistribution*) );
  for ( int i = 0; i < _stateSpaceSize; i++ ) {
    transDis[i] = _generator->getTransDistrib( i );
  }

  // prepare result and temporary tables
  res = new simulationResult( _stateSpaceSize, CONTINUOUS, stats );
  maxTrajLength = 1024;
  if ( traj ) {
    /* allocate initial memory (1k words) for trajectory */
    res->setTrajectory(true);
    dates = (double*)malloc( maxTrajLength * sizeof(double) );
    states = (int*)malloc( maxTrajLength * sizeof(int) );
    if ( withIncrements ) {
      increments = (double*)malloc( maxTrajLength * sizeof(double) );
    }
  }

  // Sample from initial distribution
  if ( (discreteDistribution*)NULL == _initDistribution ) {
    cerr << "Warning in simulateChainCT: no initial distribution. Dirac(0) assumed." << endl;
    _initDistribution = new diracDistribution(0);
  }
  curState = _initDistribution->sample();
  
  time = 0.0;

  if ( stats ) {
    cumTime = (double*) calloc( (size_t)_stateSpaceSize, sizeof( double ) );
    proba = (double*) calloc( (size_t)_stateSpaceSize, sizeof( double ) );
  }

  do {

    if ( traj || ( curTrajLength > maxTrajLength ) ) {
      maxTrajLength += 1024;
      dates = (double*)realloc( dates, maxTrajLength * sizeof(double) );
      states = (int*)realloc( states, maxTrajLength * sizeof(int) );
      if ( withIncrements ) {
	increments = (double*)realloc( increments,
				       maxTrajLength * sizeof(double) );
      }
    }
    int nextState = transDis[curState]->sample();
    double nu = - _generator->getEntry( curState, curState );
    adv = Distribution::exponential( 1.0/nu );

    if ( stats ) {
      cumTime[ curState ] += ( time+adv >= tMax ? tMax - time : adv );
    }

    if ( traj ) {

      dates[curTrajLength] = time;
      states[curTrajLength] = curState;
      if ( withIncrements ) {
	increments[curTrajLength] = adv;
      }

      if ( trace ) {
	printf("[%4d] %12.6f %4d", curTrajLength, time, curState );

	if ( withIncrements )
	  printf(" %10.6f", adv );

	printf("\n");
      }
    }

    time += adv;
    curState = nextState;

    curTrajLength++;
  } while ( time < tMax );

  /* adjust end of simulation */
  if ( traj ) {
    // By convention: last time is tMax. Have to clip the last interval.
    if ( curTrajLength > maxTrajLength ) {
      maxTrajLength += 1;
      dates = (double*)realloc( dates, maxTrajLength * sizeof(double) );
      states = (int*)realloc( states, maxTrajLength * sizeof(int) );
      if ( withIncrements ) {
	increments = (double*)realloc( increments,
				       maxTrajLength * sizeof(double) );
      }
      dates[curTrajLength] = tMax;
      states[curTrajLength] = curState;
      if ( withIncrements )
        increments[curTrajLength] = tMax-dates[curTrajLength-1];
      
      if ( trace ) {
        printf("[%4d] %12.6f %4d", curTrajLength, time, curState );

        if ( withIncrements )
          printf(" %10.6f", adv );

        printf("\n");
      }

      curTrajLength++;
    }
  }

  if ( stats ) {
    for ( int i = 0; i < _stateSpaceSize; i++ ) {
      proba[i] = cumTime[i] / tMax;
    }

    // at the moment, there is no generic state space at this level of 
    // abstract Markov Chains. So create one for the purpose.
    double* stateSpace = (double*)malloc( _stateSpaceSize * sizeof(double) );
    for ( int i = 0; i < _stateSpaceSize; i++ ) {
      stateSpace[i] = i;
    }

    discreteDistribution *dis =
      new discreteDistribution( _stateSpaceSize, stateSpace, proba );
    res->distrib( dis );

    free( cumTime );
    free( proba );
  }

  if ( traj ) {
    res->setTrajectorySize(curTrajLength);
    res->setTrajectory(dates,states);
  }

  return res;
}

simulationResult* markovChain::simulatePSI( int tMax,
                                            bool stats, bool Traj,
                                            bool print )
{
  simulationResult* res = (simulationResult*)NULL;
  char command[1024];

  switch( _type ) {

  case DISCRETE:

    if ( Traj ) {
      // write model as MARCA file

      // call to psi_alias then psi_traj
      snprintf( command, 1024, "psi_alias -i ./%s.marca -o ./%s", 
		_abstract[0].c_str(), _abstract[0].c_str() );
      cerr << "--- Executing: " << command << endl;
      system( command );
      snprintf( command, 1024, "psi_traj -i ./%s.simu -o ./%s -s %d -l %d",
		_abstract[0].c_str(), _abstract[0].c_str(),
		(int)rint(_initDistribution->sample()), tMax );
      cerr << "--- Executing: " << command << endl;
      system( command );
      res = new simulationResult( "PSI", _abstract[0], stats );

    }
    else {
      // call to psi_mcmc
      // res = simulatePSI( rint(tMax), Stats, Traj, print);

    }
    break;

  case CONTINUOUS:
    printf("Error in markovChain::simulatePSI: cannot handle continuous-time chains. No simulation.\n" );

    break;

  default:
    printf("Error in markovChain::simulatePSI: unknown Chain type (%d). No simulation.\n",
	   _type );
  }

  return res;
}

simulationResult* markovChain::simulateChain( double tMax,
					      bool Stats, bool Traj, 
                          bool withIncrements, bool print )
{
  simulationResult* res;

  switch( _type ) {
  case DISCRETE:
    res = simulateChainDT( (int)rint(tMax), Stats, Traj, print);
    break;

  case CONTINUOUS:
    res = simulateChainCT( tMax, Stats, Traj, withIncrements, print);
    break;

  default:
    printf("Error in markovChain::simulate: unknown Chain type (%d). No simulation.\n",
	   _type );
  }

  return res;
}

simulationResult* markovChain::stationaryDistributionSample( int nbSamples )
{
  char command[1024];
  simulationResult* res;

  write( "MARCA", _abstract[0].c_str() );

  // call to psi_alias then psi_sample
  snprintf( command, 1024, "psi_alias -i ./%s.marca -o ./%s",
    _abstract[0].c_str(), _abstract[0].c_str() );
  cerr << "--- Executing: " << command << endl;
  system( command );
  snprintf( command, 1024, "psi_sample -i ./%s.simu -o ./%s -d %d",
    _abstract[0].c_str(), _abstract[0].c_str(), nbSamples );
  cerr << "--- Executing: " << command << endl;
  system( command );
  res = new simulationResult( "PSI-sample", _abstract[0], /* stats? */ true );

  return res;
}


int* markovChain::simulateHittingTime( int iState, bool *hittingSet,
                                       int nbSamples, int tMax )
{
  int* res = (int*)malloc( nbSamples*sizeof(int) );
  discreteDistribution* jump = (discreteDistribution*)NULL;

  for ( int i = 0; i < nbSamples; i++ ) {
    int curState = iState;
    int hitT = 0;
    bool newState = true;

    while( !hittingSet[curState] && ( hitT < tMax ) ) {
      hitT++;

      if ( newState ) {
        // small trick to avoid re-creating the same distrib in the case
        // of a self-loop.
        if ( jump != (discreteDistribution*)NULL) delete(jump);
        jump = _generator->getTransDistrib(curState);
      }
      int nextState = jump->sample();

      newState = ( nextState != curState );
      curState = nextState;

    }
    res[i] = hitT;
  }

  return res;
}

double* markovChain::averageHittingTimeDT( bool *hitSetIndicator )
{

    // counting the number of states in the hitting set
    int nbHit = 0;
    for ( int i = 0; i < _stateSpaceSize; i++ ) {
        if ( hitSetIndicator[i] ) nbHit++;
    }

    int nbTrans = _stateSpaceSize - nbHit;

    // Building the "I-Q" matrix
    // First constructing the alias table of transient states
    int* alias = (int*)malloc( nbTrans*sizeof(int) );
    int numRow = 0;
    for ( int i = 0; i < _stateSpaceSize; i++ ) {
      if ( !hitSetIndicator[i] ) {
        alias[numRow] = i;
        numRow++;
      }
    }
    // Creating and filling the matrix at the same time
    double** ImQ = (double**) malloc( nbTrans*sizeof(double*) );
    for ( int i = 0; i < nbTrans; i++ ) {
        ImQ[i] = (double*) calloc( nbTrans, sizeof(double) );
        ImQ[i][i] = 1.0;
        for ( int j = 0; j < nbTrans; j++ ) {
          ImQ[i][j] = ImQ[i][j] - _generator->getEntry( alias[i], alias[j] );
        }
    }
    // Solving the system (I-Q).x = 1 with the Gauss method
    double *tau = (double*) malloc( nbTrans*sizeof(double) );
    for ( int i = 0; i < nbTrans; i++ ) {
      tau[i] = 1.0;
    }

#define DUMP if ( false ) { \
    for( int ii=0; ii<nbTrans; ii++ ) { \
      for( int jj=0; jj<nbTrans; jj++ ) { \
        printf(" %10f", ImQ[ii][jj] ); \
      } \
      printf("| %10f\n", tau[ii] ); \
    } \
}

    if ( _debug ) {
      printf("Init Gauss = \n");
      DUMP
    }

    for( int j=0; j<nbTrans-1; j++ ) {
        // Finding the pivot
        double Max_Val = fabs( ImQ[j][j] );
        int argMax = j;
        for( int k=j; k<nbTrans; k++ )
            if ( fabs( ImQ[k][j] ) > Max_Val ) {
                Max_Val = fabs( ImQ[k][j] );
                argMax = k;
            }
        if ( Max_Val == 0.0 ) {
            printf("Error! Singular matrix ImQ.\n");
            exit(-4);
        }
        if ( argMax != j ) {
            /* need to permute rows. beark. */
            if ( _debug )
                printf("Swapping rows %d <-> %d\n", j, argMax );
            for ( int l=j; l<nbTrans; l++ ) {
                double tmp = ImQ[j][l]; ImQ[j][l] = ImQ[argMax][l]; ImQ[argMax][l] = tmp;
            }
            double tmp = tau[j]; tau[j] = tau[argMax]; tau[argMax] = tmp;
        }
       for( int i=j+1; i<nbTrans; i++ ) {
         double ratio = -ImQ[i][j] / ImQ[j][j];
         for ( int k=j; k<nbTrans; k++ ) {
           ImQ[i][k] += ratio * ImQ[j][k];
         }
         tau[i] += ratio * tau[j];
       }
       if ( _debug ) {
         printf("Gauss/down ImQ[%d] = \n", j );
         DUMP
       }
     }
     /* Gauss phase 2: normalizing */
     for( int i=0; i<nbTrans; i++ ) {
       double rat = 1.0/ImQ[i][i];
       tau[i] *= rat;
       for ( int j=i; j<nbTrans; j++ )
         ImQ[i][j] *= rat;
     }
     if ( _debug ) {
       printf("Gauss/Matrix B/N = \n" );
       DUMP
     }
     /* Gauss phase 2: going up! */
     for( int j=nbTrans-1; j>=0; j-- ) {
       for( int i=0; i<j; i++ ) {
         tau[i] -= ImQ[i][j] * tau[j];
       }
       if ( _debug ) {
         printf("Gauss2/Matrix H[%d] =", j );
         for( int k=0; k<nbTrans; k++ )
           printf(" %f", tau[k] );
         printf("\n");
       }
     }

     // expand the tau vector into the complete hitting time vector
     double *tau_complete = (double*)calloc( _stateSpaceSize, sizeof(double) );
     for( int i = 0; i < nbTrans; i++ ) {
       tau_complete[alias[i]] = tau[i];
     }

     // cleanup
     free( tau );
     free( alias );
     for ( int i = 0; i < nbTrans; i++ ) {
       free( ImQ[i] );
     }
     free( ImQ );

     // done!
     return tau_complete;
}

double* markovChain::averageHittingTimeDT_iterative( bool *hitSetIndicator )
{
  double *tau = (double*)malloc( _stateSpaceSize * sizeof(double) );
  double *ntau = (double*)malloc( _stateSpaceSize * sizeof(double) );

  // computing an initial value for tau from a lower bound
  double minPro = 1.0;
  for ( int i = 0; i < _stateSpaceSize; i++ ) {
    if ( !hitSetIndicator[i] ) {
      double total = 0.0;
      int nbTrans = _generator->getNbElts(i);
      for ( int k = 0; k < nbTrans; k++ ) {
        if ( !hitSetIndicator[_generator->getCol(i,k)] ) {
          total += _generator->getEntryByCol(i,k);
        }
      }
      if ( _debug ) {
        fprintf( stdout, "Internal proba from state %3d: %8.6f\n", i, total );
      }

      if ( total < minPro ) minPro = total;
    }
  }
  double minTau = 1.0/(1.0-minPro);
  fprintf( stdout, "Minimal tau: %10f\n", minTau );
  for ( int i = 0; i < _stateSpaceSize; i++ ) {
    tau[i] = ( hitSetIndicator[i] ? 0.0 :minTau );
  }

  // performing iterations
  double distL1;
  int nbIter = 0;
  int maxIter = 10000;

  do {
    // evaluate P x tau
    _generator->evaluateValue( tau, ntau );

    // add the vector of ones and correct the new tau,
    // calculating L1 distance at the same time.
    distL1 = 0.0;
    for ( int i = 0; i < _stateSpaceSize; i++ ) {
      if ( hitSetIndicator[i] ) {
        ntau[i] = 0.0;
      }
      else {
        ntau[i] += 1.0;
        distL1 += fabs( tau[i] - ntau[i] );
      }
    }
    nbIter++;

    // swap tau and new tau
    double *tmp = tau; tau = ntau; ntau = tmp;

    if ( _debug ) {
      fprintf( stderr, "Hitting time iteration #%5d, distance = %12e\n", nbIter, distL1 );
    }

  } while( ( distL1 > 1e-8 ) && ( nbIter < maxIter ) );

  if ( nbIter == maxIter ) {
    fprintf( stderr, "Warning in markovChain::averageHittingTimeDT_iterative: maximum number of iterations reached. Result may be imprecise\n" );
  }

  // cleanup
  free( ntau );

  // done!
  return tau;
}

double* markovChain::averageHittingTime( bool *hitSetIndicator )
{
    double*	distRes;

    switch( _type )
    {
    case DISCRETE:
      distRes = averageHittingTimeDT( hitSetIndicator );
      break;

    case CONTINUOUS: {
      markovChain* unif = uniformize();
      distRes = unif->averageHittingTimeDT( hitSetIndicator );
      // divide by the uniformization factor

      double nu = unif->generator()->uniformizationRate();
      for ( int i = 0; i < _stateSpaceSize; i++ ) {
        distRes[i] = distRes[i] / nu;
      }

      delete unif;
    }
      break;

    default:
      printf("Error: unknown Chain type (%d). No solution.\n", _type );
    }


    return distRes;

}

void markovChain::write( FILE *out, bool withReward )
{
  if ( withReward ) {
    WARNING( "Rewards not handled. Ignored." );
  }

  if ( _type == CONTINUOUS ) {
    fprintf( out, "continuous sparse\n" );
  }
  else {
    fprintf( out, "discrete sparse\n" );
  }
  fprintf( out, "%d\n", _stateSpaceSize );

  _generator->write( out, "Ers" );
  // write reward when there is one

  fprintf( out, "stop\n" );
}

void markovChain::write(string format, string modelName)
{
  FILE* out;

  if ( format == "Ers" ) {
    string fileName = modelName + ".ers";
    out = fopen( fileName.c_str(), "w" );
    if ( NULL != out ) {
      write( out, /*reward?*/ false );

      if ( (Distribution*)NULL != _initDistribution ) {
        fprintf( out, "%d\n", (int)rint(_initDistribution->sample()) );
      }
      else {
        fprintf( out, "0\n" );
      }
      fclose( out );

    }
    else {
      fprintf( stderr, "Error: in markovChain::write(): cannot open file '%s'' for writing.\n",
               fileName.c_str() );
    }
  }
  else {
    fprintf( stderr, "Error: in markovChain::write(): format '%s' not supported.\n",
             format.c_str() );

  }
}


Distribution *  markovChain::read()
{

    Distribution * d=NULL;

if(format()=="XbornePres")
    {
      int sz=0;
      double * vals=NULL;
       double * probas=NULL;

//deserialization

      // char *mgp=(char *)(modelName().c_str());


      // mainGthLD(mgp,"Rii", &probas, &sz);


     setFormat("Xborne");

    d= new discreteDistribution(sz,vals,probas);

     }


    return d;
}


void markovChain::setSizeType(const string path)
{
  ifstream fp(path.c_str());
  double value;
  int in,ent,NbRow,NbCol,n;

  fp >> NbRow; // rows number
  fp >> NbCol; // columns number
  _stateSpaceSize = NbRow;
  fp >> n;     // non zero values number

  // read transition entries until a diagonal element is read
  bool keepReading = true;
  for(int i = 0; ( i < n ) && keepReading; i++)
    {
      fp >> in;
      fp >> ent;
      fp >> value;

      if ( in == ent ) {
	if ( value < 0.0 ) {
	  _type = CONTINUOUS;
	}
	else {
	  _type = DISCRETE;
	}
	keepReading = false;
      }
    }
  fp.close();

  cerr << "File " << path << ": found chain of size " << _stateSpaceSize
       << " and type " << ( _type == CONTINUOUS ? "continuous" : "discrete" )
       << endl;

}
