/* Marmote is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Marmote is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Marmote. If not, see <http://www.gnu.org/licenses/>.

Copyright 2015 Alain Jean-Marie, Jean-Michel Fourneau, Jean-Marc Vincent, Issam Rabhi */

#include "sparseMatrix.h"
#include <stdlib.h>
#include "../Distribution/diracDistribution.h"

/**
 * Standard constructor for sparse matrices. The internal structures
 * are initialized. The result is the null matrix.
 */
sparseMatrix::sparseMatrix(int size)
{
  _size = (long int)size;
  _debug = false;

  // allocation
  _nbElts = (int*) malloc( (size_t)_size * sizeof(int) );
  _elts = (int**) malloc( (size_t)_size * sizeof(int*) );
  _vals = (double**) malloc( (size_t)_size * sizeof(double*) );

  // initialization of contents
  if ( ( (int*)NULL != _nbElts )
       && ( (int**)NULL != _elts )
       && ( (double**)NULL != _vals ) ) {
    for ( int i=0; i<_size; i++ ) {
      _nbElts[i] = 0;
      _elts[i] = (int*)NULL;
      _vals[i] = (double*)NULL;
    }
  } 
  else {
    // abort
    if ( (int*)NULL != _nbElts ) free( _nbElts );
    if ( (int**)NULL != _elts ) free( _elts );
    if ( (double**)NULL != _vals ) free( _vals );
    // should throw some exception here
  }

}

/**
 * @brief Standard destructor
 * @author Alain Jean-Marie
 */
sparseMatrix::~sparseMatrix()
{
  if ( (int*)_nbElts != NULL ) {
    free( _nbElts );
  }
  if ( (int**)_elts != NULL ) {
    for ( int i = 0; i < _size; i++ ) free( _elts[i] );
    free( _elts );
  }
  if ( (int**)_vals != NULL ) {
    for ( int i = 0; i < _size; i++ ) free( _vals[i] );
    free( _vals );
  }
}

// Inserting an element in the matrix.
bool sparseMatrix::setEntry(int row, int col, double val)
{
  // fprintf( stderr, "R=%d C=%d V=%f\n", row, col, val );
  // checking ranges
  if ( ( 0 <= row ) && ( row < _size )
       && ( 0 <= col ) && ( col < _size ) ) {
    _nbElts[row]++;
    _elts[row] = (int*)realloc( _elts[row],
             (size_t)_nbElts[row]*sizeof(int) );
    _vals[row] = (double*)realloc( _vals[row],
            (size_t)_nbElts[row]*sizeof(double) );

    // checking that allocation was successful
    if ( ( (int*)NULL == _elts[row] )
     || ( (double*)NULL == _vals[row] ) ) {
      return false;
    }
    else {
      _elts[row][_nbElts[row]-1] = col;
      _vals[row][_nbElts[row]-1] = val;
      return true;
    }
  }
  else {
    printf("Warning: setting value to (%d,%d) in matrix of size %ld.",
       row, col, _size );
    printf(" Ignored.\n");
    return false;
  }
}

// adding a value to the matrix
bool sparseMatrix::addToEntry(int row, int col, double val)
{
  int i;

  // checking ranges
  if ( ( 0 <= row ) && ( row < _size )
       && ( 0 <= col ) && ( col < _size ) ) {
    // checking if the element is there already
    for ( i=0; 
          ( i<_nbElts[row] ) && ( _elts[row][i]!= col );
          i++ );
    if ( i<_nbElts[row] ) {
      // found it.
      _vals[row][i] += val;
      return true;
    }
    else { 
      /* not there. add element. I replicate the code of "setEntry" */
      /* for the sake of efficiency.					*/
      _nbElts[row]++;
      _elts[row] = (int*)realloc( _elts[row],
               (size_t)_nbElts[row]*sizeof(int) );
      _vals[row] = (double*)realloc( _vals[row],
              (size_t)_nbElts[row]*sizeof(double) );
      
      if ( ( (int*)NULL == _elts[row] )
       || ( (double*)NULL == _vals[row] ) ) {
	return false;
      }
      else {
    _elts[row][_nbElts[row]-1] = col;
    _vals[row][_nbElts[row]-1] = val;
	return true;
      }
    }
  }
  else {
    printf("Warning: setting value to (%d,%d) in matrix of size %ld.",
       row, col, _size );
    printf(" Ignored.\n");
    return false;
  }
}

// retrieving the entry of a matrix
double sparseMatrix::getEntry(int row, int col)
{
  double	res = 0.0;

  // the entire array has to be scanned because column index may be replicated
  for( int i=0; i<_nbElts[row]; i++ ) {
    if ( _elts[row][i] == col ) {
      res += _vals[row][i];
    }
  }

  return res;
}

// get the number of entries in row i
int sparseMatrix::getNbElts(int row)
{
  int res = 0;

  if ( ( row < 0 ) || ( row > _size ) ) {
    fprintf( stderr, "Warning in sparseMatrix::getNbElts:" );
    fprintf( stderr, " index %d out of bounds. 0 returned.\n",
	     row );
  }
  else {
    res = _nbElts[row];
  }

  return res;
}

// get the jth column for row i
int sparseMatrix::getCol(int row, int numCol)
{
  int res = -1;

  if ( ( numCol < 0 ) || ( numCol > _nbElts[row] ) ) {
    fprintf( stderr, "Warning in sparseMatrix: index %d out of bounds. -1 returned.\n",
	     numCol );
  }
  else {
    res = _elts[row][numCol];
  }

  return res;
}

// get the value in the jth column for row i
double sparseMatrix::getEntryByCol(int row, int numCol)
{
  double res = 0.0;

  if ( ( numCol < 0 ) || ( numCol > _nbElts[row] ) ) {
    fprintf( stderr, "Warning in sparseMatrix: index %d out of bounds. 0 returned.\n",
	     numCol );
  }
  else {
    res = _vals[row][numCol];
  }

  return res;
}

// transitions from some state
discreteDistribution* sparseMatrix::getTransDistrib(int state)
{
  double* next;
  double* probas;
  discreteDistribution *res;

  switch ( _type ) {
  case DISCRETE:
    // for a discrete-time structure, it is not supposed to happen
    // that the number of elements is 0.
    next = (double*) malloc( (size_t)_nbElts[state] * sizeof(double) );
    probas = (double*) malloc( (size_t)_nbElts[state] * sizeof(double) );

    for ( int j = 0; j < _nbElts[state]; j++ ) {
      next[state] = _elts[state][j];
      probas[state] = _vals[state][j];
    }
    res = new discreteDistribution( _size, next, probas );
    break;

  case CONTINUOUS:
    // for a continuous-time structure, it may happen that the number of
    // elements is 0. Since there is no discrete distribution with 0 values,
    // we force the distribution to be Dirac at i.
    if ( _nbElts == 0 ) {
      res = new diracDistribution( (double)state );
    }
    else {
      // assumption that the diagonal entry is unique in the list...
      next = (double*) malloc( (size_t)(_nbElts[state]-1) * sizeof(double) );
      probas = (double*) malloc( (size_t)(_nbElts[state]-1) * sizeof(double) );

      double totRate = getEntry(state,state);
      int idx = 0;
      for ( int j = 0; j < _nbElts[state]; j++ ) {
	if ( _elts[state][j] != state ) {
	  next[idx] = _elts[state][j];
	  probas[idx] = _vals[state][j] / totRate;
	  idx++;
	}
      }

      res = new discreteDistribution( _size, next, probas );
    }
  }

  return res;
}

// summing a row
double sparseMatrix::rowSum(int row)
{
  double sum = 0.0;

  for( int i=0; i<_nbElts[row]; i++) {
    sum += _vals[row][i];
  }

  return sum;
}

/* Linear algebra... */
// multiplication vector/matrix.
void sparseMatrix::evaluateMeasure(double* m, double *res)
{
  int		k;

  // (re) initialization
  for ( int i=0; i<_size; i++ ) 
    res[i] = 0.0;

  for ( int i=0; i<_size; i++ ) {
    for ( int j=0; j<_nbElts[i]; j++ ) {
      k = _elts[i][j];
      res[k] += m[i] * _vals[i][j];
    }
  }
}

// multiplication vector/matrix with distributions
void sparseMatrix::evaluateMeasure(discreteDistribution* d, discreteDistribution *res)
{
  int		k;
  double	sum;

  for ( int i=0; i<_size; i++ ) {
    sum = 0.0;
    for ( int j=0; j<_nbElts[i]; j++ ) {
      k = _elts[i][j];
      sum += d->getProba(k) * _vals[i][j];
    }
    res->setProba( i, sum );
  }
}

// multiplication matrix/vector
void sparseMatrix::evaluateValue(double* v, double *res)
{
  for ( int i=0; i<_size; i++ ) {
    res[i] = 0.0;
    for ( int j=0; j<_nbElts[i]; j++ ) {
      res[i] += _vals[i][j] * v[ _elts[i][j] ];
    }
  }
}

// Standard copy.
sparseMatrix* sparseMatrix::copy()
{
  sparseMatrix* res;

  res = new sparseMatrix( _size );
  res->setType( _type );

  for ( int i=0; i < _size; i++ ) {
    for ( int j=0; j < _nbElts[i]; j++ ) {
      res->setEntry( i, j, _vals[i][j] );
    }
  }

  return res;

}

// Uniformization
sparseMatrix* sparseMatrix::uniformize()
{
  sparseMatrix* res;
  double lambda;
  double nu = 0.0;

  if ( _type != CONTINUOUS ) {
    fprintf( stderr, "Error: cannot uniformize matrix. Copy of the original returned.\n" );
    res = copy();
  }
  else if ( _size == INFINITE_STATE_SPACE_SIZE ) {
    fprintf( stderr, "Error: cannot uniformize sparse matrices with infinite state spaces. Null returned.\n" );
    res = (sparseMatrix*)NULL;
  }
  else {
    // assert: continuous-time generator with finite state space.
    // Computing the uniformization factor
    for ( int i=0; i<_size; i++ ) {
      lambda = -_vals[i][i];
      nu = ( lambda > nu ? lambda : nu );
    }
    if ( _debug ) {
      printf("Uniformizing with value nu = %f\n", nu );
    }
    
    sparseMatrix* uGen = new sparseMatrix( _size );
    uGen->setType( DISCRETE );
    uGen->setUniformizationRate( nu );

    for ( int i=0; i < _size; i++ ) {
      for ( int j=0; j < _nbElts[i]; j++ ) {
	if ( _elts[i][j] != i )
	  uGen->setEntry( i, j, _vals[i][j] / nu );
	else
	  uGen->setEntry( i, j, 1.0 + _vals[i][j] / nu );
      }
    }

    fprintf( stdout, "Uniformized = \n" );
    uGen->write( stdout, "Ers" );

    res = uGen;
  }

  return res;
}

// General output method
void sparseMatrix::write( FILE* out, std::string format )
{

  if ( NULL == out ) {
    return;
  }

  // Arrays necessary for consolidating transition probabilities.
  // Consolidating is necessary because many formats do not allow
  // multiple entry declaration with the convention that they add up.
  int maxNbElts = 0;
  for ( int i = 0; i < _size; i++ )
    if ( _nbElts[i] > maxNbElts ) maxNbElts = _nbElts[i];
  int* destinations = (int*)malloc( maxNbElts*sizeof(int) );
  double* values = (double*)malloc( maxNbElts*sizeof(double) );

  if ( format == "XBORNE" ) {

#define FORMAT( ii, jj, pro ) { fprintf( out, " %12e %10d", pro, jj ); }

    for ( int i=0; i < _size; i++ ) {
      int nbTrans = consolidate( i, destinations, values );
      fprintf( out, "%10d %10d", i, nbTrans );
      for ( int d = 0; d < nbTrans; d++ ) {
        FORMAT( i, destinations[d], values[d] );
      }
      fprintf( out, "\n" );
    }
  }
  else if ( ( format == "MARCA" ) || ( format == "MatrixMarket-sparse" ) ) {
    // both MARCA and Matrix Market are "1-based"

#undef FORMAT
#define FORMAT( ii, jj, pro ) { fprintf( out, "%10d %10d %12e\n", ii+1, jj+1, pro ); }

    // enumerate all states
    for ( int i = 0; i < _size; i++ ) {
      // handle state
      int nbTrans = consolidate( i, destinations, values );
      for ( int d = 0; d < nbTrans; d++ ) {
        FORMAT( i, destinations[d], values[d] );
      }
    }
  }
  else if ( format == "Ers" ) {

#undef FORMAT
#define FORMAT( ii, jj, pro ) { fprintf( out, "%10d %10d %12e\n", ii, jj, pro ); }

    for ( int i=0; i < _size; i++ ) {
      for ( int j=0; j < _nbElts[i]; j++ ) {
        FORMAT( i, _elts[i][j], _vals[i][j] );
      }
    }
  }
  else if ( format == "Maple" ) {
    // Maple matrices are "1-based".

#undef FORMAT
#define FORMAT( ii, jj, pro ) { fprintf( out, "(%d,%d)=%e,\n", ii+1, jj+1, pro ); }
#define FORMAT2( ii, jj, pro ) { fprintf( out, "(%d,%d)=%e\n", ii+1, jj+1, pro ); }

    for ( int i = 0; i < _size; i++ ) {
      // handle state
      int nbTrans = consolidate( i, destinations, values );
      if ( i < _size-1 ) {
        for ( int d = 0; d < nbTrans; d++ ) {
          FORMAT( i, destinations[d], values[d] );
        }
      }
      else {
        for ( int d = 0; d < nbTrans-1; d++ ) {
          FORMAT( i, destinations[d], values[d] );
        }
        // avoid comma at the end of last record
        FORMAT2( i, destinations[nbTrans-1], values[nbTrans-1] );
      }
    }
  }
  else if ( format == "MatrixMarket-full" ) {
    // format is "column-major" (!)
    for ( int j=0; j<_size; j++ ) {
      for ( int i=0; i<_size; i++ ) {
        double val = getEntry( i, j );
        fprintf( out, " %12e", val );
      }
      fprintf( out, "\n");
    }
  }
  else if ( format == "Full" ) {
   for ( int i=0; i<_size; i++ ) {
     for ( int j=0; j<_size; j++ ) {
       double val = getEntry( i, j );
       fprintf( out, " %10.8f", val );
     }
     fprintf( out, "\n");
    }
  }
  else if ( format == "R" ) {
    // R format is a full matrix written as a vector
    fprintf( out, "c(");
    for ( int i=0; i<_size; i++ ) {
      for ( int j=0; j<_size; j++ ) {
        double val = getEntry( i, j );
        if ( ( j < _size-1 ) || ( i < _size-1 ) ) {
          fprintf( out, " %10.8f,", val );
        }
        else {
          fprintf( out, " %10.8f)", val );
        }
      }
    }
  }
  else if ( format == "scilab" ) {
    // Scilab sparse format. Not complete.
    fprintf( out, "sparse( [ ");
    double val;
    for ( int i=0; i<_size; i++ ) {
      for ( int j=0; j<_size-1; j++ ) {
        val = getEntry( i, j );
        fprintf( out, " %10.8f,", val );
      }
      if ( i < _size-1 ) {
        val = getEntry( i, _size-1 );
        fprintf( out, " %10.8f;", val );
      }
      else {
        fprintf( out, " %10.8f]", val );
      }
    }
  }
  else if ( format == "Matlab" ) {
    // Matlab sparse format
    // First print array of row indices
    fprintf( out, "theRows = [");
    for ( int i=0; i<_size; i++ ) {
      for ( int j=0; j<_nbElts[i]; j++ ) {
        fprintf( out, " %4d", i );
      }
    }
    fprintf( out, "]';\n" );
    // Next print array of column indices
    fprintf( out, "theColumns = [");
    for ( int i=0; i<_size; i++ ) {
      for ( int j=0; j<_nbElts[i]; j++ ) {
        fprintf( out, " %4d", _elts[i][j] );
      }
    }
    fprintf( out, "]';\n" );
    // Finally print array of values
    fprintf( out, "theValues = [");
    for ( int i=0; i<_size; i++ ) {
      for ( int j=0; j<_nbElts[i]; j++ ) {
        fprintf( out, " %12f", _vals[i][j] );
      }
    }
    fprintf( out, "]';\n" );
    // To be used with "sparse( theRows, theColumns, theValues );
  }
  else {
    fprintf( stderr, "Warning in sparseMatrix::write(): format %s not recognized. Ignored.",
             format.c_str() );
  }

  free( destinations );
  free( values );

}

// method to consolidate transitions from some state
int sparseMatrix::consolidate(int i, int* destinations, double* values)
{
  int nbTrans = 0;

  for ( int e = 0; e < _nbElts[i]; e++ ) values[e] = 0.0;

  // straighforward insertion. No sorting, no shifting.
  for ( int e = 0; e < _nbElts[i]; e++ ) {
    int dest = _elts[i][e];
    bool isNew = true;
    int location = nbTrans;
    for ( int j = 0; isNew && ( j < nbTrans ); j++ ) {
      isNew = ( dest == destinations[j] );
      location = j;
    }
    if ( isNew ) {
      destinations[location] = dest;
      values[location] = _vals[i][e];
    }
    else {
      values[location] += _vals[i][e];
    }
  }

  return nbTrans;
}
